
;
; OpenSSL primitives for Sack
; Copyright (c) 2012,2013 Mikael More
;

; Client example:
; (import sack/src/http-client openssl-sack)
; (http-get-url "https://en.wikipedia.org/")

; Server example:
(define openssl-sack-example
  '(

    (import openssl openssl-sack lib/sack/src/server sack/src/x-www-form-urlencoded)

    (define ssl-certificate-chain-file "cert.ca-bundle")
    (define ssl-certificate-file       "cert.crt"      )
    (define ssl-private-key-file       "cert.pem"      )

    (sack-start!
     (sack-app-give-x-www-form-urlencoded/UTF-8-form-post-decoder
      (lambda (env)
        (let ((ret #t))
          (values 200
                  '(("Content-Type" . "text/html"))
                  (lambda ()
                    (and
                     ret
                     (begin
                       (set! ret #f)
                       (with-output-to-u8vector
                        (list char-encoding: 'UTF-8)
                        (lambda ()
                          (write (env 'form:post))
                          (display "Hello, world!")
                          (display "<form method=post><input type=text name=field><input type=submit></form>"))))))))))
     port-number: 3333
     make-io-primitives: (let ((ssl-ctx (make-server-ssl-ctx)))

                           (if ssl-certificate-chain-file
                               (if (not (ssl-ctx-use-certificate-chain-file ssl-ctx ssl-certificate-chain-file))
                                   (error "Failed to load SSL certificate chain file" ssl-certificate-chain-file)))
                           (if ssl-certificate-file
                               (if (not (ssl-ctx-use-certificate-file ssl-ctx ssl-certificate-file OpenSSL-FileType-PEM))
                                   (error "Failed to load SSL certificate file" ssl-certificate-file)))
                           (if ssl-private-key-file
                               (if (not (ssl-ctx-use-private-key-file ssl-ctx ssl-private-key-file OpenSSL-FileType-PEM))
                                   (error "Failed to load SSL private key file" ssl-private-key-file)))

                           (ssl-ctx->make-io-primitives ssl-ctx)))

    ))

(import openssl sack/src/io-primitives sack/src/http-client)

(declare (block) (standard-bindings) (extended-bindings))

; ## Sack HTTP server support:

(define (ssl-ctx->with-io-primitives ssl-ctx #!optional (server? #t))
  (lambda (gambit-port timeout-seconds thunk)
    (let* (; threadsafe? = #f, as Sack's handling of a TCP connection is completely singlethreaded.
           ;               Gambit port timeout handlers are invoked completely asynchronously with other code though,
           ;               though this code performs timeout logic internally and is designed not to be interrupted by such.
           ; (It has one TCP connection per thread, and handles several connections simultaneously this way.)
           (channel ((if server?
                         tcp-port->ssl-server-channel
                         tcp-port->ssl-client-channel)
                     gambit-port ssl-ctx #f timeout-seconds))
           (io-primitives (make-io-primitives
                           ; read-subu8vector :
                           (channel 'read-subu8vector)
                           ; read-substring :
                           (channel 'read-substring)
                           ; read-u8 :
                           (channel 'read-u8)
                           ; write-subu8vector :
                           (channel 'write-subu8vector)
                           ; display :
                           (channel 'display)
                           ; force-output :
                           (channel 'force-output)
                           ; close-port :
                           (channel 'close-port)
                           ; read-line-until-crlf :
                           (channel 'make-read-line-until-crlf)
                           )))
      (thunk io-primitives))))

; ## Sack HTTP client support:
; Re-adding soon.
; (define (sack-http-client-add-https-support! #!optional (http-client:add-uri-scheme-io-primitives! http-client:add-uri-scheme-io-primitives!))
;   (http-client:add-uri-scheme-io-primitives! "https" (ssl-ctx->with-io-primitives client-sockets-ctx #f))) ; server? = #f i.e. client mode.
; 
; (sack-http-client-add-https-support!)
