;
; Gambit No-Exception IO module
; Copyright (C) Marc Feeley, 2012 Mikael More.
;
; Implemented for Gambit v4.6.6.
;
; Provides variants of Gambit's IO routines that do not raise exceptions, but instead return #!eof etc. .
; This module more or less directly copies rather large amounts of code out of primarily _io.scm .
;
; It would be possible to patch most of these routines within Gambit's sources into having extra arguments
; or making parameter objects that would make them deliver for this purpose. For right now though,
; Gambit's sources don't change anyhow and using this module is quite convenient.
;
; Is focused around providing routines for TCP data channel ports. Does generally not support other ports.
;
; ** Currently write-subu8vector/noexception does not work. Not clear why.
;
; Note that we explicitly do not include |display| in the list of operations in this module. This is because
; I had some weird errors from it, summarized well with this:
;      0  ##mutex-lock-out-of-line!
;      1  ##make-device-port
;      2  ##write-generic-to-character-port
;      3  ##display
;      4  io-primitives#make-standard-primitives         "lib/sack/src/io-primitives.scm"@50:1          (display 8#v 1#gambit-port)

(cond-expand
 ; Black Hole use
 (black-hole
  (export read-u8/noexception
          read-subu8vector/noexception
          read-substring/noexception
          force-output/noexception
          close-port/noexception))
 ; Vanilla Gambit use
 (else
  #!void))

(declare (block) (standard-bindings) (extended-bindings) (mostly-fixnum) (fixnum))

(define-macro (rvap idx) ; Raw vector access procedore
  `(lambda (struct) (##vector-ref struct ,idx)))

(define-macro (rvsp idx) ; Raw vector set procedore
  `(lambda (struct new-value) (##vector-set! struct ,idx new-value)))

; I don't see why the structure accessors in Gambit's sources are prefixed macro- , I suppose they're procedures anyhow.
; Now that this is the convention, to keep the code below as close to its original in the Gambit sources as possible,
; we keep them with the same names.

; _io#.scm:93
(define macro-port-mutex          (rvap 1 ))

(define macro-port-rkind          (rvap 2 ))
(define macro-port-wkind          (rvap 3 ))

(define macro-port-name           (rvap 4 ))
(define macro-port-read-datum     (rvap 5 ))
(define macro-port-write-datum    (rvap 6 ))
(define macro-port-newline        (rvap 7 ))
(define macro-port-force-output   (rvap 8 ))
(define macro-port-close          (rvap 9 ))
(define macro-port-roptions       (rvap 10))
(define macro-port-rtimeout       (rvap 11))
(define macro-port-rtimeout-thunk (rvap 12))
(define macro-port-set-rtimeout   (rvap 13))
(define macro-port-woptions       (rvap 14))
(define macro-port-wtimeout       (rvap 15))
(define macro-port-wtimeout-thunk (rvap 16))
(define macro-port-set-wtimeout   (rvap 17))

(define macro-port-mutex-set!          (rvsp 1 ))

(define macro-port-rkind-set!          (rvsp 2 ))
(define macro-port-wkind-set!          (rvsp 3 ))

(define macro-port-name-set!           (rvsp 4 ))
(define macro-port-read-datum-set!     (rvsp 5 ))
(define macro-port-write-datum-set!    (rvsp 6 ))
(define macro-port-newline-set!        (rvsp 7 ))
(define macro-port-force-output-set!   (rvsp 8 ))
(define macro-port-close-set!          (rvsp 9 ))
(define macro-port-roptions-set!       (rvsp 10))
(define macro-port-rtimeout-set!       (rvsp 11))
(define macro-port-rtimeout-thunk-set! (rvsp 12))
(define macro-port-set-rtimeout-set!   (rvsp 13))
(define macro-port-woptions-set!       (rvsp 14))
(define macro-port-wtimeout-set!       (rvsp 15))
(define macro-port-wtimeout-thunk-set! (rvsp 16))
(define macro-port-set-wtimeout-set!   (rvsp 17))

; _io#.scm:184
(define macro-character-port-rbuf             (rvap 18))
(define macro-character-port-rlo              (rvap 19))
(define macro-character-port-rhi              (rvap 20))
(define macro-character-port-rchars           (rvap 21))
(define macro-character-port-rlines           (rvap 22))
(define macro-character-port-rcurline         (rvap 23))
(define macro-character-port-rbuf-fill        (rvap 24))
(define macro-character-port-peek-eof?        (rvap 25))

(define macro-character-port-wbuf             (rvap 26))
(define macro-character-port-wlo              (rvap 27))
(define macro-character-port-whi              (rvap 28))
(define macro-character-port-wchars           (rvap 29))
(define macro-character-port-wlines           (rvap 30))
(define macro-character-port-wcurline         (rvap 31))
(define macro-character-port-wbuf-drain       (rvap 32))

(define macro-character-port-input-readtable  (rvap 33)) 
(define macro-character-port-output-readtable (rvap 34))
(define macro-character-port-output-width     (rvap 35))

(define macro-character-port-rbuf-set!             (rvsp 18))
(define macro-character-port-rlo-set!              (rvsp 19))
(define macro-character-port-rhi-set!              (rvsp 20))
(define macro-character-port-rchars-set!           (rvsp 21))
(define macro-character-port-rlines-set!           (rvsp 22))
(define macro-character-port-rcurline-set!         (rvsp 23))
(define macro-character-port-rbuf-fill-set!        (rvsp 24))
(define macro-character-port-peek-eof?-set!        (rvsp 25))

(define macro-character-port-wbuf-set!             (rvsp 26))
(define macro-character-port-wlo-set!              (rvsp 27))
(define macro-character-port-whi-set!              (rvsp 28))
(define macro-character-port-wchars-set!           (rvsp 29))
(define macro-character-port-wlines-set!           (rvsp 30))
(define macro-character-port-wcurline-set!         (rvsp 31))
(define macro-character-port-wbuf-drain-set!       (rvsp 32))

(define macro-character-port-input-readtable-set!  (rvsp 33)) 
(define macro-character-port-output-readtable-set! (rvsp 34))
(define macro-character-port-output-width-set!     (rvsp 35))

; _io#.scm:250
(define macro-byte-port-rbuf       (rvap 36))
(define macro-byte-port-rlo        (rvap 37))
(define macro-byte-port-rhi        (rvap 38))
(define macro-byte-port-rbuf-fill  (rvap 39))

(define macro-byte-port-wbuf       (rvap 40))
(define macro-byte-port-wlo        (rvap 41))
(define macro-byte-port-whi        (rvap 42))
(define macro-byte-port-wbuf-drain (rvap 43))

(define macro-byte-port-rbuf-set!       (rvsp 36))
(define macro-byte-port-rlo-set!        (rvsp 37))
(define macro-byte-port-rhi-set!        (rvsp 38))
(define macro-byte-port-rbuf-fill-set!  (rvsp 39))

(define macro-byte-port-wbuf-set!       (rvsp 40))
(define macro-byte-port-wlo-set!        (rvsp 41))
(define macro-byte-port-whi-set!        (rvsp 42))
(define macro-byte-port-wbuf-drain-set! (rvsp 43))

; _io#.scm:277
(define macro-device-port-rdevice-condvar (rvap 44))
(define macro-device-port-wdevice-condvar (rvap 45))
(define macro-device-port-name            (rvap 46))

(define macro-device-port-rdevice-condvar-set! (rvsp 44))
(define macro-device-port-wdevice-condvar-set! (rvsp 45))
(define macro-device-port-name-set!            (rvsp 46))

; _io#.scm:646
(define-macro (macro-port-mutex-lock! port)
  `(macro-mutex-lock! (macro-port-mutex ,port) #f (macro-current-thread)))

(define-macro (macro-port-mutex-unlock! port)
  `(macro-mutex-unlock! (macro-port-mutex ,port)))

; ** The original macro-mutex* were included only to make write-subu8vector work, which it does not, so these can be reverted.

; Really this one is at _thread#.scm:1302 though right now this is fine -
; Nope did not work for write-subu8vector.
; (define (macro-mutex-lock! mutex absrel-timeout owner)
;   (mutex-lock! mutex))

; _thread#.scm:1302
(define-macro (macro-mutex-lock! mutex absrel-timeout owner)
  `(let ((mutex ,mutex) (absrel-timeout ,absrel-timeout) (owner ,owner))

     (declare (not interrupts-enabled))

     (let ((state (macro-btq-owner mutex)))
       (if state
         (##mutex-lock-out-of-line! mutex absrel-timeout owner)
         (begin
           (macro-btq-link! mutex owner)
           #t)))))

; (define (macro-mutex-unlock! mutex)
;   (mutex-unlock! mutex))

(define-macro (macro-mutex-unlock! mutex)
  `(let ((mutex ,mutex))

     (declare (not interrupts-enabled))

     (macro-btq-deq-remove! mutex)
     (let ((leftmost (macro-btq-leftmost mutex)))
       (if (##eq? leftmost mutex)
         (begin
           (macro-btq-unlink! mutex #f)
           (##void))
         (##mutex-signal! mutex leftmost #f)))))

; ** These were included only to make write-subu8vector work, which it does not, so these can be reverted.
; _thread#.scm:851:
(define-macro (macro-btq-deq-init! deq)
  `(let ((deq ,deq))

     (declare (not interrupts-enabled))

     (macro-btq-deq-next-set! deq deq)
     (macro-btq-deq-prev-set! deq deq)))

(define-macro (macro-btq-deq-remove! item)
  `(let ((item ,item))

     (declare (not interrupts-enabled))

     (let ((item-next (macro-btq-deq-next item))
           (item-prev (macro-btq-deq-prev item)))
       (macro-btq-deq-next-set! item-prev item-next)
       (macro-btq-deq-prev-set! item-next item-prev))))

(define-macro (macro-btq-deq-insert! deq item)
  `(let ((deq ,deq) (item ,item))

     (declare (not interrupts-enabled))

     ;; add item to tail of deq

     (let ((deq-last (macro-btq-deq-prev deq)))
       (macro-btq-deq-next-set! deq-last item)
       (macro-btq-deq-prev-set! deq item)
       (macro-btq-deq-next-set! item deq)
       (macro-btq-deq-prev-set! item deq-last))))

(define-macro (macro-btq-link! mutex thread)
  `(let ((mutex ,mutex) (thread ,thread))

     (declare (not interrupts-enabled))

     (macro-btq-deq-insert! thread mutex)
     (macro-btq-owner-set! mutex thread)))

(define-macro (macro-btq-unlink! mutex state)
  `(let ((mutex ,mutex) (state ,state))

     (declare (not interrupts-enabled))

     (macro-btq-deq-init! mutex)
     (macro-btq-owner-set! mutex state)))

; ** These were included only to make write-subu8vector work, which it does not, so these can be reverted.
; _thread#.scm:833:
;;; Representation of blocked thread queues.
(define-macro (macro-btq-deq-next node)        `(macro-slot 1 ,node))
(define-macro (macro-btq-deq-next-set! node x) `(macro-slot 1 ,node ,x))
(define-macro (macro-btq-deq-prev node)        `(macro-slot 2 ,node))
(define-macro (macro-btq-deq-prev-set! node x) `(macro-slot 2 ,node ,x))
(define-macro (macro-btq-color node)           `(macro-slot 3 ,node))
(define-macro (macro-btq-color-set! node x)    `(macro-slot 3 ,node ,x))
(define-macro (macro-btq-parent node)          `(macro-slot 4 ,node))
(define-macro (macro-btq-parent-set! node x)   `(macro-slot 4 ,node ,x))
(define-macro (macro-btq-left node)            `(macro-slot 5 ,node))
(define-macro (macro-btq-left-set! node x)     `(macro-slot 5 ,node ,x))
(define-macro (macro-btq-right node)           `(macro-slot 6 ,node))
(define-macro (macro-btq-right-set! node x)    `(macro-slot 6 ,node ,x))
(define-macro (macro-btq-leftmost node)        `(macro-slot 6 ,node))
(define-macro (macro-btq-leftmost-set! node x) `(macro-slot 6 ,node ,x))
(define-macro (macro-btq-owner node)           `(macro-slot 7 ,node))
(define-macro (macro-btq-owner-set! node x)    `(macro-slot 7 ,node ,x))

; ** This was included only to make write-subu8vector work, which it does not, so these can be reverted.
; Wild guess:
(define-macro (macro-slot index object #!optional v)
  (if v
      `(##vector-set! ,object ,index ,v)
      `(##vector-ref ,object ,index)))

; _thread#.scm:1357
; 
(define macro-condition-variable-btq-deq-next (rvap 1))
(define macro-condition-variable-btq-deq-prev (rvap 2))
(define macro-condition-variable-btq-color    (rvap 3))
(define macro-condition-variable-btq-parent   (rvap 4))
(define macro-condition-variable-btq-left     (rvap 5))
(define macro-condition-variable-btq-leftmost (rvap 6))
(define macro-condition-variable-btq-owner    (rvap 7))
(define macro-condvar-name                    (rvap 8))
(define macro-condvar-specific                (rvap 9))

(define macro-condition-variable-btq-deq-next-set! (rvap 1))
(define macro-condition-variable-btq-deq-prev-set! (rvap 2))
(define macro-condition-variable-btq-color-set!    (rvap 3))
(define macro-condition-variable-btq-parent-set!   (rvap 4))
(define macro-condition-variable-btq-left-set!     (rvap 5))
(define macro-condition-variable-btq-leftmost-set! (rvap 6))
(define macro-condition-variable-btq-owner-set!    (rvap 7))
(define macro-condvar-name-set!                    (rvsp 8))
(define macro-condvar-specific-set!                (rvsp 9))

; Improvised
(define-macro (macro-current-thread) `(current-thread))

; _io#.scm:141
(define-macro (macro-none-kind)      0) ;; allows nothing

; _io#.scm:765
(define-macro (macro-unbuffered? options)
  `(##fixnum.< (##fixnum.bitwise-and ,options 2047) 1024))

; Deps:
(define-macro (macro-lock-and-check-input-port-character-buffer-empty
               port
               form
               expr)
  `(begin

     (macro-port-mutex-lock! ,port) ;; get exclusive access to port

     (if (or (##fixnum.< (macro-character-port-rlo ,port)
                         (macro-character-port-rhi ,port))
             (macro-character-port-peek-eof? ,port))

       (begin
         (macro-port-mutex-unlock! ,port)
         ; (##raise-nonempty-input-port-character-buffer-exception ,@form)
         'nonempty-input-port-character-buffer
         )

       ,expr)))



(define (read-u8/noexception port)

  (declare (not interrupts-enabled))

  (macro-lock-and-check-input-port-character-buffer-empty
   port
   (read-u8 port)
   (let loop ()
     (let* ((byte-rlo
             (macro-byte-port-rlo port))
            (byte-rhi
             (macro-byte-port-rhi port)))
       (if (##fixnum.< byte-rlo byte-rhi)
           (let* ((byte-rbuf
                   (macro-byte-port-rbuf port))
                  (result
                   (##u8vector-ref byte-rbuf byte-rlo)))
             (macro-byte-port-rlo-set! port (##fixnum.+ byte-rlo 1))
             (macro-port-mutex-unlock! port)
             result)
           (let ((code
                  ((macro-byte-port-rbuf-fill port)
                   port
                   1
                   #t)))
             (cond ((##fixnum? code)

                    ;; an error occurred

                    (macro-port-mutex-unlock! port)

                    (if (##fixnum.= code ##err-code-EAGAIN)
                        #!eof ;; the read timeout thunk returned #f
                        ; (##raise-os-exception
                        ;  #f
                        ;  code
                        ;  read-u8
                        ;  port)
                        #!eof
                        ))

                   (code

                    ;; bytes were added to byte buffer, so try again
                    ;; to transfer bytes from the byte buffer

                    (loop))

                   (else

                    ;; no bytes were added to byte buffer
                    ;; (end-of-file was reached)

                    (macro-port-mutex-unlock! port)
                    #!eof))))))))

(define (close-port-or-force-output/noexception operation)

  ; _io.scm:1455
  ; In there this procedure goes straight as the close argument to |macro-make-device-port|,
  ; meaning that it's this procedure that ##close-port invokes directly. Thus the arguments are:
  ; port close-port port (close-port as in a reference to the primitive that's being invoked.
  ; I believe it's passed around as to refer to in case of exceptions.)
  (define (close port prim arg1)

    ;; It is assumed that the thread **does not** have exclusive
    ;; access to the port.

    (declare (not interrupts-enabled))

    (macro-port-mutex-lock! port) ;; get exclusive access to port

    (let ((result (close-aux1 port prim)))
      (macro-port-mutex-unlock! port)
      (if (##fixnum? result)
          (begin
            ; (##raise-os-exception #f result prim arg1)
            #!eof)
          result)))

  (define (close-aux1 port prim)

    ;; It is assumed that the thread has exclusive access to the port.

    (declare (not interrupts-enabled))

    (if (or (##fixnum.= (macro-port-wkind port) (macro-none-kind))
            (##eq? prim close-input-port))
        (close-aux2 port prim)
        (let ((code (force-output-aux port 0 #f)))
          (if (and (##fixnum.< code 0)
                   (##not (##fixnum.= code ##err-code-EAGAIN)))

              code

              ;; The close operation may have failed to force the output.
              ;; However the close operation is not allowed to block, so
              ;; we just continue and close the device.  The user can make
              ;; sure that the output is forced by calling force-output
              ;; (which can block) before calling close-port.

              (close-aux2 port prim)))))

  (define (close-aux2 port prim)

    ;; It is assumed that the thread has exclusive access to the port.

    (declare (not interrupts-enabled))

    (##close-device
       port
       (macro-device-port-rdevice-condvar port)
       (macro-device-port-wdevice-condvar port)
       prim))

  ; _io.scm:1392
  ; For inspiration, here's how ##force-output feeds these args:
  ; (define-prim (##force-output
  ;               port
  ;               #!optional
  ;               (level (macro-absent-obj)))
  ;   (##declare (not interrupts-enabled))
  ;   ((macro-port-force-output port)
  ;    port
  ;    (if (##eq? level (macro-absent-obj)) 0 level)
  ;    force-output
  ;    port
  ;    level
  ;    (macro-absent-obj)
  ;    (macro-absent-obj)))
  (define (force-output port level prim arg1 arg2 arg3 arg4)

    ;; It is assumed that the thread **does not** have exclusive
    ;; access to the port.

    (declare (not interrupts-enabled))

    (macro-port-mutex-lock! port) ;; get exclusive access to port

    (let ((code (force-output-aux port level #t)))
      (macro-port-mutex-unlock! port)
      (if (##fixnum.< code 0)
          (begin
            ; (##raise-os-exception #f code prim arg1 arg2 arg3 arg4)
            #!eof)
          (##void))))

  ; _io.scm:1407
  (define (force-output-aux port level block?)

    ;; It is assumed that the thread has exclusive access to the port.

    (declare (not interrupts-enabled))

    (let ((code1 (drain-output port)))
      (if (##fixnum? code1)
          code1
          (let* ((wdevice-condvar (macro-device-port-wdevice-condvar port))
                 (wdevice (macro-condvar-name wdevice-condvar))
                 (code2 (##os-device-force-output wdevice level)))
            (cond ((##fixnum.= code2 ##err-code-EINTR)

                   ;; the force was interrupted, so try again

                   (force-output-aux port level block?))

                  ((and block?
                        (##fixnum.= code2 ##err-code-EAGAIN))

                   ;; the force would block, so wait and then try again

                   (macro-port-mutex-unlock! port)
                   (let ((continue?
                          (or (##wait-for-io!
                                 (macro-device-port-wdevice-condvar port)
                                 (macro-port-wtimeout port))
                              ((macro-port-wtimeout-thunk port)))))
                     (macro-port-mutex-lock! port) ;; regain access to port
                     (if continue?
                         (force-output-aux port level block?)
                         code2)))

                  (else
                   code2))))))

  (define (drain-output port)

    ;; It is assumed that the thread has exclusive access to the port.

    (declare (not interrupts-enabled))

    (let ((code ((macro-character-port-wbuf-drain port) port)))
      (if (##fixnum? code)
          code
          ((macro-byte-port-wbuf-drain port) port))))

  (case operation
    ((close-port)
     (lambda (port)
       (close port
              #f ; We don't use the prim argument here anyhow.
              port)))
    ((force-output)
     (lambda (port level)
       (force-output port
                     level
                     #f ; prim not needed
                     port
                     level
                     #f
                     #f)))))

; Aliasing this one with ##close-port did not deliver.
(define (close-port/noexception port)
   ((close-port-or-force-output/noexception 'close-port) port))

(define (force-output/noexception port #!optional (level 0))
  ((close-port-or-force-output/noexception 'force-output) port level))

(define (read-subu8vector/noexception u8vect
                                      start
                                      end
                                      port
                                      #!optional
                                      (need 1))

  (declare (not interrupts-enabled))

  (macro-lock-and-check-input-port-character-buffer-empty
   port
   (read-subu8vector u8vect start end port need)
   (let loop ((n 0))
     (let ((remaining (##fixnum.- end (##fixnum.+ start n))))
       (if (##not (##fixnum.< 0 remaining))
           (begin
             (macro-port-mutex-unlock! port)
             n)
           (let* ((byte-rlo
                   (macro-byte-port-rlo port))
                  (byte-rhi
                   (macro-byte-port-rhi port))
                  (bytes-buffered
                   (##fixnum.- byte-rhi byte-rlo)))
             (if (##fixnum.< 0 bytes-buffered)

                 (let* ((to-transfer
                         (##fixnum.min remaining bytes-buffered))
                        (limit
                         (##fixnum.+ byte-rlo to-transfer))
                        (byte-rbuf
                         (macro-byte-port-rbuf port)))
                   (macro-byte-port-rlo-set! port limit)
                   (##subu8vector-move!
                    byte-rbuf
                    byte-rlo
                    limit
                    u8vect
                    (##fixnum.+ start n))
                   (loop (##fixnum.+ n to-transfer)))

                 (let ((code
                        ((macro-byte-port-rbuf-fill port)
                         port
                         remaining
                         (or (##not (##fixnum? need))
                             (##fixnum.< n need)))))
                   (cond ((##fixnum? code)

                          ;; an error occurred, signal an error if no
                          ;; bytes were previously transferred from byte
                          ;; buffer and (in the case of a read timeout)
                          ;; the timeout thunk returned #f

                          (macro-port-mutex-unlock! port)

                          (if (or (##fixnum.< 0 n)
                                  (##fixnum.= code ##err-code-EAGAIN))
                              n
                              (begin
                                ; (##raise-os-exception #f code read-subu8vector u8vect start end port need)
                                #!eof)
                              ))

                         (code

                          ;; bytes were added to byte buffer, so try again
                          ;; to transfer bytes from the byte buffer

                          (loop n))

                         (else

                          ;; no bytes were added to byte buffer
                          ;; (end-of-file was reached)

                          (macro-port-mutex-unlock! port)
                          n))))))))))

(define (read-substring/noexception str
                                    start
                                    end
                                    port
                                    #!optional
                                    (need 1))

  (declare (not interrupts-enabled))

  (let loop ((n 0))
    (let ((remaining (##fixnum.- end (##fixnum.+ start n))))
      (if (##not (##fixnum.< 0 remaining))
          (begin
            (macro-port-mutex-unlock! port)
            n)
          (let* ((char-rlo
                  (macro-character-port-rlo port))
                 (char-rhi
                  (macro-character-port-rhi port))
                 (chars-buffered
                  (##fixnum.- char-rhi char-rlo)))
            (if (##fixnum.< 0 chars-buffered)

                (let* ((to-transfer
                        (##fixnum.min remaining chars-buffered))
                       (limit
                        (##fixnum.+ char-rlo to-transfer))
                       (char-rbuf
                        (macro-character-port-rbuf port)))
                  (macro-character-port-rlo-set! port limit)
                  (##substring-move!
                   char-rbuf
                   char-rlo
                   limit
                   str
                   (##fixnum.+ start n))
                  (let loop2 ((rlo char-rlo))
                    (if (##fixnum.< rlo limit)
                        (let ((c (##string-ref char-rbuf rlo))
                              (rlo+1 (##fixnum.+ rlo 1)))

                          (if (##char=? c #\newline)
                              (begin

                                ;; keep track of number of characters read

                                (let ((char-rchars
                                       (macro-character-port-rchars port)))
                                  (macro-character-port-rcurline-set! port
                                    (##fixnum.+ char-rchars rlo+1)))

                                ;; keep track of number of lines read

                                (let ((char-rlines
                                       (macro-character-port-rlines port)))
                                  (macro-character-port-rlines-set! port
                                    (##fixnum.+ char-rlines 1)))))

                          (loop2 rlo+1))))
                  (loop (##fixnum.+ n to-transfer)))

                (let ((code
                       ((macro-character-port-rbuf-fill port)
                        port
                        remaining
                        (or (##not (##fixnum? need))
                            (##fixnum.< n need)))))
                  (cond ((##fixnum? code)

                         ;; an error occurred, signal an error if no
                         ;; chars were previously transferred from char
                         ;; buffer and (in the case of a read timeout)
                         ;; the timeout thunk returned #f

                         (macro-port-mutex-unlock! port)

                         (if (or (##fixnum.< 0 n)
                                 (##fixnum.= code ##err-code-EAGAIN))
                             n
                             (begin
                               ; (##raise-os-exception #f code read-substring str start end port need)
                               #!eof)
                             ))

                        (code

                         ;; chars were added to char buffer, so try again
                         ;; to transfer chars from the char buffer

                         (loop n))

                        (else

                         ;; no chars were added to char buffer
                         ;; (end-of-file was reached)

                         (macro-port-mutex-unlock! port)
                         n)))))))))

; DOES NOT WORK. Not clear about why.
(define (write-subu8vector/noexception u8vect start end port)

  (declare (not interrupts-enabled))

  (macro-port-mutex-lock! port) ;; get exclusive access to port

  (let ((code
         (and (##fixnum.< (macro-character-port-wlo port)
                          (macro-character-port-whi port))
              ((macro-character-port-wbuf-drain port) port))))
    (if (##fixnum? code)

      (begin
        (macro-port-mutex-unlock! port)
        (if (##fixnum.= code ##err-code-EAGAIN)
          0
          (begin
            ; (##raise-os-exception #f code write-subu8vector u8vect start end port)
            #!eof)))

      (let loop1 ((n 0))
        (let ((remaining (##fixnum.- end (##fixnum.+ start n))))
          (if (##not (##fixnum.< 0 remaining))
            (begin

              ;; force output if port is set for unbuffered output

              (if (and (##fixnum.< start end)
                       (macro-unbuffered? (macro-port-woptions port)))
                (begin
                  (macro-port-mutex-unlock! port)
                  (begin
                    ; Might throw exception:
                    ; Note that all the arguments from write-subu8vector and on are actually ignored by force-output-
                    ; ((macro-port-force-output port) port 0 write-subu8vector u8vect start end port)
                    ; Exception-safe:
                    (force-output/noexception port 0)
                    )
                (macro-port-mutex-unlock! port))

              n)
            (let* ((byte-whi
                    (macro-byte-port-whi port))
                   (byte-wbuf
                    (macro-byte-port-wbuf port))
                   (bytes-free
                    (##fixnum.- (##u8vector-length byte-wbuf) byte-whi)))
              (if (##fixnum.< 0 bytes-free)
                (let* ((to-transfer
                        (##fixnum.min remaining bytes-free))
                       (limit
                        (##fixnum.+ byte-whi to-transfer)))
                  (macro-byte-port-whi-set! port limit)
                  (let loop2 ((i (##fixnum.+ start n))
                              (j byte-whi))
                    (if (##fixnum.< j limit)
                      (begin
                        (##u8vector-set! byte-wbuf j (##u8vector-ref u8vect i))
                        (loop2 (##fixnum.+ i 1)
                               (##fixnum.+ j 1)))))
                  (loop1 (##fixnum.+ n to-transfer)))
                (let ((code ((macro-byte-port-wbuf-drain port) port)))
                  (if (##fixnum? code)
                    (begin

                      ;; an error occurred, signal an error if no bytes
                      ;; were previously transferred from byte buffer
                      ;; and (in the case of a write timeout) the
                      ;; timeout thunk returned #f

                      (macro-port-mutex-unlock! port)

                      (if (or (##fixnum.< 0 n)
                              (##fixnum.= code ##err-code-EAGAIN))
                        n
                        (begin
                          ; (##raise-os-exception #f code write-subu8vector u8vect start end port)
                          #!eof)))

                    ;; the byte buffer was successfully drained, so try
                    ;; again to transfer bytes to the byte buffer

                    (loop1 n))))))))))))

; We had a display implementation here too, did not work, more info about that in the top header.
; Not clear about why it not worked. We need only a minimum implementation of display that's
; easily simulated using write-subu8vector though, so being without it is fine.








; Notes from the research that led to the making of this library
; 
; # Source in Gambit's code of the "Device is closed" error message
; Tracking down the "Device is closed" message that |read-u8| outputs when a port is or becomes closed, that it is trying to read from:
;
; It's generated by the ___os_err_code_to_string procedure, based on the ___CLOSED_DEVICE_ERR error code.
; 
; This procedure is reachable as ##os-err-code->string .
; 
; The place in the code that this simply must come from is ___device_tcp_client_read_raw_virt (os_io.c:3893) and
; there's a corresponding ___device_tcp_client_write_raw_virt (os.io:3938).
; 
; The only place in all code that these two procedures are invoked from are through their references in the
; ___device_tcp_client_vtbl structure by the name ___device_tcp_client_table (os.io:4025).
; 
; This type is an alias of the ___device_stream_vtbl type.
; 
; (There's a corresponding ___device_tcp_server_vtbl type for servers (os.io.c:4425).)
; 
; Its first element ___device_tcp_client_kind returns ___TCP_CLIENT_DEVICE_KIND .
; 
; In turn ___device_stream_vtbl_struct is a ___device_vtbl (os_io.h:470).
; 
; 
; ##read-u8 is defined at _io.scm:4939. 
; 
; Spontaneously it looks like it throws its errors based on a fixnum return value from its (macro-byte-port-rbuf-fill port) call (_io.scm:4960).
; 
; rbuff-fill of byte_port is defined at _io#.scm:250, it's the fourth slot.
; 
; It's defined using define-type-of-character-port , character-port is defined at _io#.scm:184. In it, rbuf-fill is the seventh slot.
; 
; character-port is a type of port, port is defined at _io#.scm:93. It has no rbuf-fill.
; 
; 
; The tcp-client objects' ##vector-ref 0 say #<type #4 device-port> . device-port is defined at _io#.scm row 277, it's a type of byte port.
; It has no rbuf-fill.
; 
; 
; Presuming the rbuf-fill we're talking about is the one of byte-port, it's slot number:
;  * 1 for structure type
;  * 17 for all port slots
;  * (there's a rbuf-fill at slot number 7 in the character port)
;    18 for all character port slots
;  * 4:th slot
; 
; So, index 1 + 17 + 18 + 4 - 1 = 39
; 
; The char port thing is at 1 + 17 + 7 - 1 = 24
; 
; Looking at them in Gambit:
; 
; 
; (##vector-ref c 39) => #<procedure #2 ##byte-rbuf-fill>
; (##vector-ref c 24) => #<procedure #3 ##char-rbuf-fill>
; 
; Makes perfect sense - so, slot 39 is byte-rbuf-fill.
; 
; And..
; 
; ((##vector-ref c 39) c 1 #t) => -507510778
; (##os-err-code->string #) => "Device is closed"
; 
; Yes! :D
; 
; 
; # Closing a socket port
; 
; This is performed through the reference to ___device_tcp_client_close_raw_virt in the ___device_tcp_client_table structure (os_io.c:1025).
; 
; It's not invoked directly but ___device_stream_close_raw_virt which is an alias is invoked by ___device_stream_close_virt (os_io.c:1794).
; 
; ___device_stream_close_virt is not invoked ever though.
; 
; 
; The alias of ___device_vtbl (os_io.h:470) has a an alias for close by the name ___device_close_virt .
; 
; This one is essentially aliased by ___device_close .
; 
; ___device_close in turn is invoked by ___device_cleanup and ___os_device_close 
; 
; ___os_device_close is aliased by ##os-device-close (_io.scm:1637)
; 
; ##os-device-close is invoked by ##close-device .
; 
; 
; ##close-device is invoked by ##make-device-port (a huge procedure!) ##make-tcp-server-port ##make-directory-psettings ##make-event-queue-port
; 
; Here we go: The closing is done by close, close-aux1 and close-aux2 (_io.scm:1455).
;


